#!/usr/bin/env python

"""This file contains physical constants."""

# Avogadro number
N_avogadro = 6.022140857e23
# Planck's constant
h_Planck = 6.62607004e-34
# Boltzmann constant J/K
kB = 1.38064852e-23
# Atomic mass unit. kg
amu = 1.66053886e-27
